// ===============================================
// File:		GLSettings.cxx
// Description:		setup the OpenGL state machine
// Current:             Sept 28th 1999
// Author(s):		Thomas Kern
// ===============================================

/*
 *        thk VRML-Viewer (based on OpenGL)
 *  it shows static *.wrl files based on the VRML97 standard
 *
 *          Copyright (C) 1997-1999  Thomas Kern
 *                  All Rights Reserved
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * My email adress is: 
 *   Thomas.Kern@3Dsource.de 
 * 
 * You should find these files at: my homepage
 *   http://www.3Dsource.de
 *
 */


//=====================================================================
//	INCLUDES
//=====================================================================
//
#include <math.h>
#include <stdio.h>

#include "GLSettings.H"


//=====================================================================
//	GLOBALS
//=====================================================================
//
static GLdouble  ViewingAngle, WindowAspect, nearPlane, farPlane;
static GLfloat   eyeM[16], projM[16], modelM[16];
static GLfloat   calcedViewM[16], calcedObjM[16];
static GLfloat   moveVect[3], curEyeDist;
static GLfloat   L0pos[4], fogStart, fogEnd;
static GLfloat   rotX, rotY, rotZ;
static GLclampf  L0amb[4], L0diff[4], fogColor[4], clearColor[4];

static void CalcEyeMatrix(); // forward declaration


//=====================================================================
//	STEUERUNG DES OPENGL-STATUS
//=====================================================================
//
// Initialisierung durch eine Datei
static void InitFromFile()
{

}

// --------------------------------------------------------------------
// Perspektive korrigieren
// ViewingAngle: field of view angle(degrees), in y direction.
// WindowAspect: aspect ratio = field of view in x direction. 
// The aspect ratio is the ratio of x (width) to y (height).
void GLSetPerspective(int newWidth, int newHeight)
{ 
  if( newWidth && newHeight ) {
    // 1: Berechnungen zur Neufestlegung durchfuehren
    WindowAspect = (GLfloat)newWidth / (GLfloat)newHeight;
    // 2: Viewport neu festlegen
    glViewport(0, 0, newWidth, newHeight);
  }
  // 3: die Projektionsmatrix veraendern
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  // 4: das Blickfeld neu festlegen
  gluPerspective(ViewingAngle, WindowAspect, 
		             nearPlane,    farPlane);

  // 5: die berechnete Projektionsmatrix abrufen
  glGetFloatv(GL_PROJECTION_MATRIX, projM);

  // 6: alle folgenden Operationen beeinflussen die Modellmatrix
  glMatrixMode(GL_MODELVIEW);
  CalcEyeMatrix();
}

// --------------------------------------------------------------------
// Festlegen der Startwerte, wird durch die Config-datei ueberschrieben 
void GLReset(int currentWidth, int currentHeight)
{
  int i;

  ViewingAngle    =  15.0; // degrees
  nearPlane       =   0.5;
  farPlane        = 500.0;

  fogStart   =   10.0; // in [m] vom Augpunkt
  fogEnd     =  500.0;

  // die Matrizen alle auf Null, dann nur einzelne Felder setzen
  for(i = 0; i < 3; i++) {
    moveVect[i] =  0.0;
    L0pos[i]    = 10.0;
    L0amb[i]    =  0.15;
    L0diff[i]   =  0.85;
    fogColor[i] =  0.85;
    clearColor[i] = 0.5;
  }
  moveVect[1] =  -1.0;
  moveVect[2] = -10.0;
  L0pos[3] = 0.0;
  L0amb[3] = L0diff[3] = 1.0;
  fogColor[2] = 0.93; fogColor[3] = 1.0; 
  clearColor[3] = 1.0;

  rotX = rotY = rotZ = 0.0;

  // andere Vorgabewerte aus einer Datei einlesen
  // InitFromFile();

  // hier noch die OpenGL Staten setzten
  glClearColor(clearColor[0], clearColor[1], clearColor[2], clearColor[3]);    
  glEnable(GL_DEPTH_TEST);

  // die zuletzt gespeicherten Matrizen loeschen
  for(i = 0; i < 16; i++) {
    modelM[i] = projM[i] = eyeM[i] = 0.0;
    calcedObjM[i] = calcedViewM[i] = 0.0;
  }
  // die Augpunkt-Matrix muss als Einheitsmatrix definiert sein:
  eyeM[ 0] = 1.0;
  eyeM[ 5] = 1.0;
  eyeM[10] = 1.0;
  eyeM[15] = 1.0;  

  // erst jetzt die Berechnungen durchfuehren
  GLSetPerspective(currentWidth, currentHeight);
}

//=====================================================================
// Funktionen zum Zeichnen und Bildraum berechnen
//=====================================================================
//
// --------------------------------------------------------------------
//
void coordDir(GLfloat xp[3], GLfloat zp[3])
{
  GLfloat p[3];

  multMatrix(calcedViewM, projM, eyeM);

  // nun die Bildgrenzen ermitteln
  // xp = viewingM * p;
  p[0] = 1.0; // spaeter noch mit ViewingAngle
  p[1] = p[2] = 0.0;
  
  xp[0] = calcedViewM[0]*p[0]; // p1,p2  sind Null
  xp[1] = calcedViewM[4]*p[0]; // + viewingM[5]*p[1] + viewingM[ 6]*p[2]; 
  xp[2] = calcedViewM[8]*p[0]; // + viewingM[9]*p[1] + viewingM[10]*p[2];

  // zp = viewingM * p;
  p[2] = 1.0;
  p[0] = p[1] = 0.0;
  
  zp[0] = calcedViewM[ 2]*p[2]; // p0,p1 sind Null
  zp[1] = calcedViewM[ 6]*p[2]; // + viewingM[4]*p[0] + viewingM[5]*p[1]
  zp[2] = calcedViewM[10]*p[2]; // + viewingM[8]*p[0] + viewingM[9]*p[1]  
}

// --------------------------------------------------------------------
// Die Bewegungen des Auges werden hier von Hand gesetzt.
void GLnewFrame()
{
  glClearColor(0.6,0.6,0.6,1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();

  // der Augpunkt ist im Ursprung des WC verankert. 
  // Blickrichtung entlang der negativen Z-Achse.
  // VUP entspricht der Y-Achse.
  glMultMatrixf(eyeM);
  // glMultMatrix ersetzt die folgenden 4 Aufrufe
  //glTranslatef(moveVect[0], moveVect[1], moveVect[2]);
  //glRotatef(rotX, 1.0, 0.0, 0.0);
  //glRotatef(rotY, 0.0, 1.0, 0.0);
  //glRotatef(rotZ, 0.0, 0.0, 1.0);
}

//=====================================================================
//	Veraendern des Matrix-Stacks
//=====================================================================
//
//
static void CalcEyeMatrix()
{
  // Die folgenden Berechnungen ermitteln die gesamte 
  // Augpunkt-Transformation, so dass beim Zeichnen nur ein
  // glMultMatrix() Aufruf ausreicht.

  // Die Translation erfolgt vor der Rotation. 
  // Damit ergibt sich die Gesamtmatrix Mg = T:
  //      |a0  a4  a8   a12|   |1  0  0  dx|
  //      |a1  a5  a9   a13|   |0  1  0  dy|
  // Mg = |a2  a6  a10  a14| = |0  0  1  dz|
  //      |a3  a7  a11  a15|   |0  0  0   1|
  
  // Dann die Rotationsmatrix gemaess:
  // Rm = R3 * (R2 * (R1 *T)) (Reihenfolge also T->R1->R2->R3)
  // mit R1 = Rx, R2 = Ry, R3 = Rz.
  //      |r0  r4  r8  r12|
  //      |r1  r5  r9  r13|
  // Rm = |r2  r6  r10 r14|
  //      |r3  r7  r11 r15|
  // 
  // mit:
  // R[0] = cosYcosZ                    
  // R[1] = sinXsinYcosZ + cosXsinZ      
  // R[2] = sinXsinZ     - cosXsinYcosZ  
  // R[3] =              - cosYsinZ
  // R[4] = cosXcosZ     - sinXsinYsinZ 
  // R[5] = cosXsinYsinZ - sinXcosZ      
  // R[6] = sinY
  // R[7] =              - sinXcosY
  // R[8] = cosXcosY
  // 
  // Die Translation erfolgt vor der Rotation, benoetigt aber 
  // eine 4*4 Matrix. Damit ergibt sich die Gesamtmatrix:
  // Mg = Rm * T; wobei Rm als R[16] interpretiert wird.
  // Die Berechnung ist allerdings trivial:
  //      |a0  a4  a8   a12|   |r0  r3  r6  dx|
  //      |a1  a5  a9   a13|   |r1  r4  r7  dy|
  // Mg = |a2  a6  a10  a14| = |r2  r5  r8  dz|
  //      |a3  a7  a11  a15|   | 0   0   0   1|
  
  GLfloat sinX, cosX, sinY, cosY, sinZ, cosZ;

  if(rotX) {
    sinX = sin(rad(rotX)); // Angabe in Radiant
    cosX = cos(rad(rotX));
  }
  else {
    sinX = 0.0; 
    cosX = 1.0;
  }
  if(rotY) {
    sinY = sin(rad(rotY)); 
    cosY = cos(rad(rotY));
  }
  else {
    sinY = 0.0; 
    cosY = 1.0;
  }
  if(rotZ) {
    sinZ = sin(rad(rotZ)); 
    cosZ = cos(rad(rotZ));
  }
  else {
    sinZ = 0.0; 
    cosZ = 1.0;
  }

  eyeM[ 0] =  cosY*cosZ;
  eyeM[ 1] = (sinX*sinY*cosZ) + (cosX*     sinZ);
  eyeM[ 2] = (sinX*     sinZ) - (cosX*sinY*cosZ);
  eyeM[ 3] = 0.0;
  eyeM[ 4] =                  - (cosY*     sinZ);
  eyeM[ 5] = (cosX*     cosZ) - (sinX*sinY*sinZ);
  eyeM[ 6] = (cosX*sinY*sinZ) - (sinX*     cosZ); 
  eyeM[ 7] = 0.0;
  eyeM[ 8] =  sinY;
  eyeM[ 9] =                  - (sinX*cosY     );
  eyeM[10] =  cosX*cosY;
  eyeM[11] = 0.0;
  eyeM[12] = moveVect[0];
  eyeM[13] = moveVect[1];
  eyeM[14] = moveVect[2];
  eyeM[15] = 1.0;  
}

// --------------------------------------------------------------------
// die Rotationsmatrix
void GLAddEyeRot(GLfloat addX, GLfloat addY, GLfloat addZ)
{
  rotX += addX;
  rotY += addY;
  rotZ += addZ;

  while(rotX > 360.0)
    rotX -= 360.0;
  while(rotX < -360.0)
    rotX += 360.0;
  while(rotY > 360.0)
    rotY -= 360.0;
  while(rotY < -360.0)
    rotY += 360.0;
  while(rotZ > 360.0)
    rotZ -= 360.0;
  while(rotZ < -360.0)
    rotZ += 360.0;
  
  CalcEyeMatrix();
}

// --------------------------------------------------------------------
// die Rotationsmatrix
void GLSetEyeRot(GLfloat angleX, GLfloat angleY, GLfloat angleZ)
{
  rotX = angleX;
  rotY = angleY;
  rotZ = angleZ;

  GLAddEyeRot(0.0, 0.0, 0.0);
}

// --------------------------------------------------------------------
//
void GLAddEyeMove(GLfloat addZ)
{
  moveVect[2] -= addZ;
  eyeM[12] = moveVect[0]; // erspart Aufruf von calcEyeMatrix();
  eyeM[13] = moveVect[1];
  eyeM[14] = moveVect[2];
  
  // neue Laenge des Blickrichtungsvektors bestimmen und speichern
  curEyeDist = sqrt( moveVect[0]*moveVect[0] + moveVect[1]*moveVect[1] + moveVect[2]*moveVect[2] );
  
  // Nebelstaerke anpassen
  glFogf(GL_FOG_START, curEyeDist +  fogStart);
  glFogf(GL_FOG_END,   curEyeDist +  fogEnd);
}

// --------------------------------------------------------------------
//
void GLSetEyePoint(GLfloat x, GLfloat y, GLfloat z)
{
  moveVect[0] = -x;
  moveVect[1] = -y;
  moveVect[2] = -z;

  GLAddEyeMove(0.0); // Laenge des Vektors berechnen
}

// --------------------------------------------------------------------
//
void GLSetPlanes(GLdouble nearP, GLdouble farP)
{
  if(nearP) // passing NULL changes nothing
    nearPlane = nearP;
  if(farP)
    farPlane = farP;
}

// --------------------------------------------------------------------
//
void GLSetAngle(GLdouble angle)
{
  if(angle)
    ViewingAngle = angle;
}


//=====================================================================
//	Veraendern von Farbe etc.
//=====================================================================
//
// Licht
void GLSetLighting(boolean SetON)
{
  if(SetON) {
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, L0amb);
    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
    //glColorMaterial(GL_FRONT,GL_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);
    glEnable(GL_LIGHTING);

    glLightfv(GL_LIGHT0, GL_AMBIENT,  L0amb);
    glLightfv(GL_LIGHT0, GL_DIFFUSE,  L0diff);
    //glLightfv(GL_LIGHT0, GL_SPECULAR, L0spec);
    glLightfv(GL_LIGHT0, GL_POSITION, L0pos);
    glEnable(GL_LIGHT0);
    //printf("  %s: \t\t %s\n", M_LIGHT, M_ON);
  }
  else {
    glDisable(GL_LIGHTING);
    glDisable(GL_COLOR_MATERIAL);
    //printf("  %s: \t\t\t %s\n", M_LIGHT, M_OFF);
  }
}

//--------------------------------------------------------------------
// Wireframe
void GLSetDrawStile(boolean SetON)
{
  if(SetON) {
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    //printf("  %s: \t\t %s\n", M_STILE, M_FILL);
  }
  else {
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    //printf("  %s: \t\t\t %s\n", M_STILE, M_LINE);
  }
}

//--------------------------------------------------------------------
// Culling
void GLSetCulling(boolean SetON)
{
  if(SetON) {
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    //printf("  %s: \t %s\n", M_CULL, M_ON);
  }
  else {
    glDisable(GL_CULL_FACE);
    //printf("  %s: \t\t %s\n", M_CULL, M_OFF);
  }
}

//--------------------------------------------------------------------
// Shading
void GLSetShading(boolean SetON)
{
  if(SetON) {
    glShadeModel(GL_SMOOTH);
    //printf("  %s: \t %s\n", M_SMOOTH, M_ON);
  }
  else {
    glShadeModel(GL_FLAT);
    //printf("  %s: \t\t %s\n", M_SMOOTH, M_OFF);
  }
}

//--------------------------------------------------------------------
// Nebel
void GLSetFog(boolean SetON)
{
  if(SetON) {
    glEnable(GL_FOG);
    glFogf(GL_FOG_MODE,   GL_LINEAR);
    glFogfv(GL_FOG_COLOR, fogColor);
  }
  else {
    glDisable(GL_FOG);
  }
}


//=====================================================================
//	Clipping von Objekten berechnen
//=====================================================================
//
//--------------------------------------------------------------------
// 
#include "VRMLControl.H"

void GLCalcObjMatrix(Quadrants Quads[2])
{
  GLfloat zA[3], viewAt[3];

  // die aktuellen Transformationen berechnen
  //     |a0  a4  a8   a12|
  //     |a1  a5  a9   a13|
  // M = |a2  a6  a10  a14|
  //     |a3  a7  a11  a15|
  // calcedObjM = projM * modelM;
  glGetFloatv(GL_MODELVIEW_MATRIX, modelM);

  multMatrix(calcedObjM, projM, modelM);

  // aktuelle Blickrichtung bestimmen (ist bei Beginn parallel zur z-Achse)
  zA[0] = zA[2] = 0.0;
  zA[1] = 1.0;
  multVector(viewAt, calcedObjM, zA);


  // DEBUG
  sprintf(myMessageString, "Check: %2.2f, %2.2f, %2.2f\t\t",
          viewAt[0], viewAt[1], viewAt[2]);
  writeMessage();
  // END DEBUG


  // welche Quadranten muessen angezeigt werden ?
  // dazu Projektion in die X,Z Ebene durch normalisieren
  normalizeVect(viewAt);    

  // nun kanns losgehen
  if( viewAt[0] > 0.0 ) {          // rechts der Z-Achse
    if( viewAt[2] < 0.0 ) {
      if( viewAt[2] > -M_SQRT_2 ) { // Quadrant I, untere Haelfte
	Quads[0] = q1u;
	Quads[1] = q4o;
      }
      else {                       // Quadrant I, obere Haelfte
	Quads[0] = q1o;
	Quads[1] = q1u;
      }
    }
    else {
      if( viewAt[2] < M_SQRT_2 ) { // Quadrant IV, obere Haelfte
	Quads[0] = q1u;
	Quads[1] = q4o;
      }
      else {                       // Quadrant IV, untere Haelfte 
	Quads[0] = q4o;
	Quads[1] = q4u;	
      }
    }
  }
  else {                           // links der Z-Achse
    if( viewAt[2] < 0.0 ) {
      if( viewAt[2] > -M_SQRT_2 ) { // Quadrant II, untere Haelfte
	Quads[0] = q3o;
	Quads[1] = q2u;
      }
      else {                       // Quadrant II, obere Haelfte
	Quads[0] = q2o;
	Quads[1] = q1o;
      }
    }
    else {
      if( viewAt[2] < M_SQRT_2 ) { // Quadrant III, obere Haelfte
	Quads[0] = q2u;
	Quads[1] = q3o;
      }
      else {                       // Quadrant III, untere Haelfte 
	Quads[0] = q4u;
	Quads[1] = q3u;	
      }
    }
  }
}

//--------------------------------------------------------------------
// 
boolean GLCheckVisible(GLfloat oldP[3])
{
  boolean visible = true;
  GLfloat newP[3];

  multVector(newP, calcedObjM, oldP);

  sprintf(myMessageString, "P[3]: %2.2f, %2.2f, %2.2f\n", newP[0], newP[1], newP[2]);
  writeMessage();

  return visible;
}
